-- FLASH interface
-- every communication between FPGA and external SRAM should use this interface

library IEEE;
use IEEE.std_logic_1164.all; 
use IEEE.std_logic_unsigned.all;

--library VIRTEX;
--use VIRTEX.all; -- for xilinx component signals


entity flash is
  generic (adr_width: integer:= 19; -- width of the address bus
    data_width: integer:= 16;  -- width of the data bus
    C_CLOCK_PERIOD_NS: integer:= 40;
    C_ACCESS_TIME_NS: integer:= 70;
    C_WRITE_PULSE_NS: integer:= 35;     
    adr_lsb_size: integer:= 1 -- the lsb address wich are generated localy = log2(c_dwidth/c_flash_dwidth)
    	); 
    port ( -- user signals
	    clk: in STD_LOGIC; -- clock signal - every user signal is latched rising edge of clk
		  arst: in std_logic; -- global asynchronous set reset 
      rnw: in STD_LOGIC; -- output enable, write enable (input to flip flops)
		  adri: in STD_LOGIC_VECTOR (0 to adr_width-1); -- address from which data will be read or written
      adr1_lsb: out std_logic_vector(0 to adr_lsb_size); -- (bit 0 is never used) the lsb address wich are generated localy = log2(c_dwidth/c_flash_dwidth)
		  datar: out STD_LOGIC_VECTOR (0 to data_width-1); -- data read form the sram
      dataw: in STD_LOGIC_VECTOR (0 to data_width-1); -- data to be written to the sram
		  ce_adr: in std_logic; -- clock enable for address flip-flops (except the LSB which always always '1')
		  stb: in std_logic; -- clock enable for datar flip-flops
      ack: out std_logic; -- flash is ready for the next write read operation (this delay is due to the access time
      retry: out std_logic; -- flash is busy on programming so this cycle should be repeated
      -- external SRAM signals
      adr: out STD_LOGIC_VECTOR (0 to adr_width-1); -- address bus 
		  data_i: in STD_LOGIC_VECTOR (0 to data_width-1); -- data bus (input)
		  data_o: out STD_LOGIC_VECTOR (0 to data_width-1); -- data bus (output)
		  data_t: out STD_LOGIC; -- data bus (tri-state)
      oen: out std_logic; -- output enable (active low)
      wen: out std_logic; -- write enable
      cen: out std_logic; -- chip enable (always zero)
      busyn: in std_logic -- Flash is busy
    );
end flash;

architecture flash_arch of flash is
 function MAX(arg0: integer; arg1: integer) return integer is
    variable ret: integer;
  begin
    if arg0>=arg1 then ret:= arg0;
    else ret:= arg1;
    end if;
    return ret;
  end MAX;
  constant read_clk: integer:= 1 + (C_ACCESS_TIME_NS + C_CLOCK_PERIOD_NS - 1)/C_CLOCK_PERIOD_NS; -- 1 + due to additioanl clock cycle to latch input data to input flifp- flops,
  constant write_clk: integer:= 1 + (C_WRITE_PULSE_NS + C_CLOCK_PERIOD_NS - 1)/C_CLOCK_PERIOD_NS; -- 1 + (due to that write pulse is generate at the falling edge of the clock
  constant access_clk: integer:= MAX(read_clk, write_clk); -- maximum number of clock cycles for a single read or write operation
  signal count: integer range 0 to access_clk; -- counter to count access time delay, 0- idle, 1.. operation cycles
  signal wr_count_last, rd_count_last, count_last, count_zero: std_logic; -- counter is = 0
  signal datawq: std_logic_vector(data_width-1 downto 0); -- dataw in the next clk cycle
  signal rnwq, ack1, busynq, rd_seq_adr: std_logic; -- read sequancial addressing is used to read the next address (only within one full OPB transfer)
  signal adr_lsb, adr_lsb_1: std_logic_vector(0 to MAX(0, adr_lsb_size-1)); -- LSB of the address which is increamented locally
  
begin
-- address flip-flops (in IOB) and multiplexer
process (clk, arst) begin
	if arst= '1' then
	  adr<= (others=>'0');
    if adr_lsb_size>0 then adr_lsb<= (others=>'0'); end if;
  elsif clk'event and clk='1' then -- insert flip-flops (FFs) without reset (output FFs)
    if ce_adr='1' and stb='1' and (count_zero='1' or (rnwq='0' and wr_count_last='1') or -- idle cycle or the last cycle
        (adr_lsb_size>0 and rd_seq_adr='1')) then -- the seqAdd for reading was assumed
         adr<= adri; -- all address bits are lathed
         if adr_lsb_size>0 then
           adr_lsb<= adri(adr_width - adr_lsb_size to adr_width-1);
         end if;
    
    elsif adr_lsb_size>0 and count_last='1' then
      adr(adr_width - adr_lsb_size to adr_width-1)<= adr_lsb_1;
      adr_lsb<= adr_lsb_1;
	  end if;
  end if;
end process;



ga: if adr_lsb_size>0 generate
  adr_lsb_1<= adr_lsb + 1;
  adr1_lsb(1 to adr_lsb_size)<= adr_lsb;
end generate;
  adr1_lsb(0)<= '0'; -- always zero - this zero iss used e.g. for k=1 when 

ack1<= '0' when busynq='0' or stb='0' else -- FLASH is not ready to accept
   count_last or (count_zero and not rnw); -- the last cycle or ( idle cycle and write)
ack<= ack1;
   
cnt: process(arst, clk) begin
  if arst='1' then count<= 0; rd_seq_adr<= '0';
  elsif clk'event and clk='1' then
    if rnwq='0' then -- writing process
      if wr_count_last='1' then
        if stb='1' then -- the next cyckle in progress
          count<= 1;
        else
          count<= 0; -- go to the idle cycle
        end if;
      else
        count<= count + 1;
      end if;
    
    else -- rnwq='1' reading or idle
      if stb='0' then -- go to idle
        count<= 0; rd_seq_adr<= '0';
      elsif rd_count_last = '1' then
        if adr_lsb_size>0 and adr_lsb /= (2**adr_lsb_size)-1 then -- use sequancial adressing (only within one full transfer on OPB)
          count<= 1; rd_seq_adr<='1'; -- read the next data in the read cycle
        else
          count<= 0; rd_seq_adr<='0';
        end if;
      elsif rd_seq_adr='1' and adr_lsb_size>0 and ce_adr='1' then -- assumed seqencial addresssing but ce_adr indicates that this is not seq adr
        count<= 1; rd_seq_adr<= '0';
      else
        count<= count + 1;
      end if;
    end if;
  end if; -- clk
end process;        
             
rd_count_last<= '1' when count = read_clk else '0';
wr_count_last<= '1' when count = write_clk else '0';
count_last<= rd_count_last when rnwq='1' else wr_count_last;
count_zero<= '1' when count = 0 else '0';

-- data path flip-flops insertet in IOB
process (clk, arst) begin
    if arst='1' then
		  data_o<= (others=> '0');
		  datar<= (others=>'0');
	elsif clk'event and clk='1' then -- insert flip-flops (FFs) without reset (output FFs)
    if ack1='1' then
		  data_o<= dataw;
		end if;
    if count = (read_clk - 1)  then
      datar<= data_i;
    end if;
	end if;
end process;
  
  -- control signal logic
  cen<= '0'; -- SRAM is always selected
  
  -- flash control signals
process (clk, arst) begin
	if arst='1' then
		oen<= '0'; rnwq<= '1'; busynq<= '0';
  elsif clk'event and clk='1' then
    if ack1='1' and rnw='0' then 
      rnwq<= '0';
      oen<= '1';
    elsif wr_count_last='1' then
      rnwq<= '1';
      oen<= '0';
    end if;
    busynq<= busyn;
	end if;
end process;
retry<= not busynq;

  -- WEN flip-flop
  process (clk, arst) begin
    if arst='1' then wen<= '1';
    elsif clk'event and clk='0' then -- falling EDGE !!!!!!!!
      wen<= rnwq or wr_count_last;
    end if;
  end process;
  
  data_t<= rnwq;
  
    
end flash_arch;
