-- SRAM to OPB block
-- author: Erenst Jamro


library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;



entity flash2opb is
  generic (
  	C_AWIDTH: integer:= 18; -- number of address bits of the OPB bus connected to the SRAM= C_SRAM_AWIDTH - log2(C_DWIDTH/C_SRAM_DWIDTH)
    C_DWIDTH : integer:= 32; -- real width of the OPB data -bus  the rest MSBs are always 0s
    adr_lsb_size: integer; -- = log2(c_dwidth / c_flash_dwidth)
	  C_FLASH_DWIDTH: integer:= 16; -- external memory data width
	  C_FLASH_AWIDTH: integer:= 19; -- external memory address width
    C_CLOCK_PERIOD_NS: integer:= 40;  
    C_WRITE_PULSE_NS: integer:= 35;     
    C_ACCESS_TIME_NS: integer:= 70);
  port (
  -- Global signals
    arst: in std_logic;
    clk : in std_logic;
    ABus    : in std_logic_vector(0 to c_awidth-1);
    DBus    : in std_logic_vector(0 to C_DWIDTH-1);
    Sl_DBus : out std_logic_vector(0 to C_DWIDTH-1);
	  -- internal signals diffrent from OPB
	  be: in std_logic_vector(0 to c_dwidth/8 -1); -- byte enable
    rnw, stb: in std_logic;
	  ack: buffer std_logic; -- TRANSFER acknowledge
    retry: out std_logic; -- retry for OPB generated only when busyn= '0'
	  -- external SRAM signals
    adr: out STD_LOGIC_VECTOR (0 to c_flash_awidth-1); -- address bus
    data_i: in STD_LOGIC_VECTOR (0 to c_flash_dwidth-1); -- data bus
	  data_o: out STD_LOGIC_VECTOR (0 to c_flash_dwidth-1); -- data bus
	  data_t: out STD_LOGIC; -- tri-state datap_o
    oen: out std_logic; -- output enable (active low)
    wen: out std_logic; -- write enable
    cen: out std_logic; -- chip enable (always zero)	
    busyn: in std_logic
    );
end flash2opb;

architecture arch of flash2opb is
   function MAX(arg0: integer; arg1: integer) return integer is
    variable ret: integer;
  begin
    if arg0>=arg1 then ret:= arg0;
    else ret:= arg1;
    end if;
    return ret;
  end MAX;

  component flash
   generic (adr_width: integer:= 19; -- width of the address bus
    data_width: integer:= 16;  -- width of the data bus
    C_CLOCK_PERIOD_NS: integer:= 40;
    C_ACCESS_TIME_NS: integer:= 70;     
    C_WRITE_PULSE_NS: integer:= 35;
    adr_lsb_size: integer:= 1 -- the lsb address wich are generated localy = log2(c_dwidth/c_flash_dwidth)
    	); 
   port ( -- user signals
	    clk: in STD_LOGIC; -- clock signal - every user signal is latched rising edge of clk
		  arst: in std_logic; -- global asynchronous set reset 
      rnw: in STD_LOGIC; -- output enable, write enable (input to flip flops)
		  adri: in STD_LOGIC_VECTOR (0 to adr_width-1); -- address from which data will be read or written
      adr1_lsb: out std_logic_vector(0 to adr_lsb_size); -- (bit 0 is never used) the lsb address wich are generated localy = log2(c_dwidth/c_flash_dwidth)
		  datar: out STD_LOGIC_VECTOR (0 to data_width-1); -- data read form the sram
      dataw: in STD_LOGIC_VECTOR (0 to data_width-1); -- data to be written to the sram
		  ce_adr: in std_logic; -- clock enable for address flip-flops (except the LSB which always always '1')
		  stb: in std_logic; -- clock enable for datar flip-flops
      ack: out std_logic; -- flash is ready for the next write read operation (this delay is due to the access time
      retry: out std_logic; -- flash is busy on programming so this cycle should be repeated
      -- external SRAM signals
      adr: out STD_LOGIC_VECTOR (0 to adr_width-1); -- address bus 
		  data_i: in STD_LOGIC_VECTOR (0 to data_width-1); -- data bus (input)
		  data_o: out STD_LOGIC_VECTOR (0 to data_width-1); -- data bus (output)
		  data_t: out STD_LOGIC; -- data bus (tri-state)
      oen: out std_logic; -- output enable (active low)
      wen: out std_logic; -- write enable
      cen: out std_logic; -- chip enable (always zero)
      busyn: in std_logic -- Flash is busy
    );
   end component;
  
  constant k: integer:= c_dwidth/c_flash_dwidth; -- number of sram cycles for one OPB cycle
  signal buf_wr, buf_rd: std_logic_vector(0 to MAX(0 , c_dwidth-c_flash_dwidth-1));
  signal be_wr: std_logic_vector(0 to k-1); -- valid write buffer; 
  signal be_rd: std_logic_vector(0 to MAX(0, k-2)); -- valid buffer specified for every c_flash_dwidth-bits (MSBs are omitted)
  signal frnw, fstb, fack, ce_adr, be_msb, be_lsb: std_logic;
  signal fadr: std_logic_vector(0 to c_flash_awidth-1);
  signal adr_lsb: std_logic_vector(0 to adr_lsb_size); -- bit (0) is always zero
  signal fdrd, fdwr: std_logic_vector(0 to c_flash_dwidth-1); -- flash data read and write
  
begin  
fl: flash
  generic map (adr_width=> c_flash_awidth, data_width=> c_flash_dwidth, C_WRITE_PULSE_NS=> C_WRITE_PULSE_NS,
    C_CLOCK_PERIOD_NS=> C_CLOCK_PERIOD_NS, C_ACCESS_TIME_NS=> C_ACCESS_TIME_NS, adr_lsb_size=> adr_lsb_size)
  port map(  --user signals
	    clk=> clk, arst=> arst, 
      rnw=> frnw, adri=> fadr, adr1_lsb=> adr_lsb, 
		  datar=> fdrd, dataw=> fdwr, ce_adr=> ce_adr,
		  stb=> fstb, ack=> fack, retry=> retry,
      -- external SRAM signals
      adr=> adr, data_i=> data_i, data_o=> data_o, data_t=>data_t,
      oen=> oen, wen=> wen, cen=> cen, busyn=> busyn);


   k2: if k=2 generate
     fstb<= '1' when be_wr(1)='1' else -- write operation in progress
       stb when rnw='1' or be(0 to c_flash_dwidth/8 -1)/= 0 else '0'; -- do not generate write operation when only LSBs write
     frnw<= '0' when be_wr(1)= '1' else rnw; -- write operation in progress
     fdwr<=  buf_wr when  be_wr(1) = '1' else
        DBus(0 to c_flash_dwidth-1);
     Sl_DBus<= buf_rd & fdrd;
     ack<= '0' when stb='0' else
           not be_wr(0) and fack when be_wr(1) = '1' else -- write buffer valid so ack='1' only when fack='1' and LSBs are not transfered
           fack when rnw='0' else -- write operation
           fack when be_msb='0' else -- read only LSBs 
           be_rd(0) when be_lsb='0' else -- read only MSBs, be_rd(0) - valid data in the MSB buffer
           fack and be_rd(0); -- read full word 
     ce_adr<= (not be_wr(1) or not be_wr(0)) and not be_rd(0); -- ce_adr logic is should be '1' only for the first transfer
     be_msb<= '1' when be(0 to c_flash_dwidth/8-1) /= 0 else '0';
     be_lsb<= '1' when be(c_flash_dwidth/8 to c_dwidth/8 -1) /= 0 else '0';
     
     process (arst, clk) begin
       if arst='1' then 
         be_wr<= "00"; buf_wr<= (others=>'0'); 
         be_rd(0)<= '0'; buf_rd<= (others=>'0');
       elsif clk'event and clk='1' then
         if rnw='0' and stb='1' and be_wr(1)='0' and (be_msb='0' or fack='1') then
           be_wr(1)<= be_lsb;  -- be_wr(0 to 1)= "x0" - buffer not valid, ="11"- full transfer (MSB already transfered), ="01" only LSBs transfer
           be_wr(0)<= be_msb;  -- be_wr(0) <= 1 only when full transfer
           buf_wr<= DBus(c_flash_dwidth to c_dwidth-1);
         elsif fack='1' then -- the write to the LSBs
           be_wr(0 to 1)<= "00";
         end if;
         
         if frnw='1' and fack='1' and adr_lsb(1)='0' then -- reading the MSBs
           be_rd(0)<= '1';
           buf_rd<= fdrd;
         elsif ack='1' then -- sending the read data out
           be_rd(0)<= '0';
         end if;
       end if;
     end process;               
   end generate;


   k1: if k=1 generate
      frnw<= rnw;  Sl_DBus<= fdrd; fdwr<= DBus; ce_adr<= '1';
      fstb<= stb; ack<= fack;
    end generate;


    fadr(c_flash_awidth - c_awidth to c_flash_awidth-1)<= ABus;
    ga: if c_flash_awidth - c_awidth > 0 generate
      fadr(0 to c_flash_awidth - c_awidth-1)<= (others=>'0');
    end generate;
    

end arch;