-- Author:          Ernest Jamro
-- OPB bus connected to the external SRAM memory (SRAM is 16 bit wide)

--//////////////////////////////////////////////////////////////////////
--//// Copyright (C) 2003 Authors                                   //// 
--////                                                              ////
--//// This source file may be used and distributed without         ////
--/// restriction provided that this copyright statement is not     ////
--//// removed from the file and that any derivative work contains  ////
--//// the original copyright notice and the associated disclaimer. ////
--////                                                              ////
--//// This source file is free software; you can redistribute it   ////
--//// and/or modify it under the terms of the GNU Lesser General   ////
--//// Public License as published by the Free Software Foundation; ////
--//// either version 2.1 of the License, or (at your option) any   ////
--//// later version.                                               ////
--////                                                              ////
--//// This source is distributed in the hope that it will be       ////
--//// useful, but WITHOUT ANY WARRANTY; without even the implied   ////
--//// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ////
--//// PURPOSE. See the GNU Lesser General Public License for more  ////
--//// details.                                                     ////
--//////////////////////////////////////////////////////////////////////

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

library UNISIM;
use unisim.all; -- for xilinx component signals


entity OPB_FLASH is
  generic (
    C_OPB_AWIDTH : integer                   := 24;
    C_OPB_DWIDTH : integer                   := 32;
    C_BASEADDR   : std_logic_vector(0 to 31) := X"0000_0000";
    C_HIGHADDR   : std_logic_vector(0 to 31) := X"0000_01FF";
    C_SPEEDRATE  : integer := 0;
    C_REAL_DWIDTH : integer:= 32; -- real width of the data -bus  the rst LSBs are always 0s
    C_FLASH_DWIDTH: integer:= 16; -- external memory data width
    C_FLASH_AWIDTH: integer:= 18; -- external memory address width
    C_CLOCK_PERIOD_NS: integer:= 40;
    C_ACCESS_TIME_NS: integer:= 70;
    C_WRITE_PULSE_NS: integer:= 35
    );
  port (
    -- Global signals
    OPB_Clk : in std_logic;
    OPB_Rst : in std_logic; -- not used in the project
    -- OPB signals
    OPB_ABus    : in std_logic_vector(0 to C_OPB_AWIDTH-1);
    OPB_BE      : in std_logic_vector(0 to C_OPB_DWIDTH/8-1);
    OPB_RNW     : in std_logic;
    OPB_select  : in std_logic;
    OPB_seqAddr : in std_logic;
    OPB_DBus    : in std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sl_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sl_errAck  : out std_logic;
    Sl_retry   : out std_logic;
    Sl_toutSup : out std_logic;
    Sl_xferAck : out std_logic;
    -- external SRAM signals
    a: out STD_LOGIC_VECTOR (0 to c_flash_awidth-1); -- address bus
    d_i: in STD_LOGIC_VECTOR (0 to c_flash_dwidth-1); -- data bus
    d_o: out STD_LOGIC_VECTOR (0 to c_flash_dwidth-1); -- data bus
    d_t: out STD_LOGIC; -- tri-state datap_o
    oen: out std_logic; -- output enable (active low)
    wen: out std_logic; -- write enable
    cen: out std_logic; -- chip enable (always zero)
    busyn: in std_logic -- flash is busy with programming
    );
end entity OPB_FLASH;

architecture arch of opb_flash is
  -------------------------------------------------------------------------
  -- functions and components declarations
  function log2(arg: integer) return integer is
    variable ret: integer:= -1;
  begin
    if arg= 0 then
      ret:= -1;
      return ret;
    end if; 
    for i in 0 to 32 loop
      if 2**i = arg then
        ret:= i;
        return i;
      end if;
    end loop;
    assert false -- always assert when loop does not finish the function
    report "opb_sram. Log2() argument is not a power of 2"
    severity failure;
    return -1;
  end log2;
  
  component ROC -- reset on Configuration
    port (O : out std_logic);
  end component;
  component flash2opb
    generic (
      C_AWIDTH: integer:= 18; -- number of address bits of the OPB bus connected to the SRAM= C_SRAM_AWIDTH - log2(C_DWIDTH/C_SRAM_DWIDTH)
      C_DWIDTH : integer:= 32; -- real width of the OPB data -bus  the rest MSBs are always 0s
      adr_lsb_size: integer; -- = log2(c_dwidth / c_flash_dwidth)
      C_FLASH_DWIDTH: integer:= 16; -- external memory data width
      C_FLASH_AWIDTH: integer:= 19; -- external memory address width
      C_CLOCK_PERIOD_NS: integer:= 40;
      C_WRITE_PULSE_NS: integer:= 35;
      C_ACCESS_TIME_NS: integer:= 70);
    port (
      -- Global signals
      arst: in std_logic;
      clk : in std_logic;
      ABus    : in std_logic_vector(0 to c_awidth-1);
      DBus    : in std_logic_vector(0 to C_DWIDTH-1);
      Sl_DBus : out std_logic_vector(0 to C_DWIDTH-1);
      -- internal signals diffrent from OPB
      be: in std_logic_vector(0 to c_dwidth/8 -1); -- byte enable
      rnw, stb: in std_logic;
      ack: buffer std_logic; -- TRANSFER acknowledge
      retry: out std_logic; -- retry for OPB generated only when busyn= '0'
      -- external SRAM signals
      adr: out STD_LOGIC_VECTOR (0 to c_flash_awidth-1); -- address bus
      data_i: in STD_LOGIC_VECTOR (0 to c_flash_dwidth-1); -- data bus
      data_o: out STD_LOGIC_VECTOR (0 to c_flash_dwidth-1); -- data bus
      data_t: out STD_LOGIC; -- tri-state datap_o
      oen: out std_logic; -- output enable (active low)
      wen: out std_logic; -- write enable
      cen: out std_logic; -- chip enable (always zero)	
      busyn: in std_logic
      );
  end component;
  component OPB_slave
    generic (
      C_OPB_AWIDTH : integer                   := 32;
      C_OPB_DWIDTH : integer                   := 32;
      C_REAL_DWIDTH: integer:= 0; -- real OPB bus data width used during data transfer (=0 - same as dwidth)
      DWIDTH : integer:= 8; -- user width of the data - the rest LSBs are always 0s
      AWIDTH : integer:= 8; -- user address width
      C_BASEADDR   : std_logic_vector          := X"0000_0100";
      C_HIGHADDR   : std_logic_vector          := X"0000_01FF";
      C_SPEEDRATE : integer := 0; -- the lower the speed the quicker (lower number of clock cycles - latency) 
      -- and the lower occupied area however the lower clock frequency is required. 
      -- Therefore to increase the clock frequency this parameter should be increased
      -- user interface data and address width
      fifo_wr_awidth: integer:= 0; -- write fifo address width (recommended value = 4) - the size of the fifo= 2^fifo_wr_awidth, =1 - only a single flip-flop; =0 direct connection- fifo is not used, <0 - no connection at all (for write logic)
      fifo_rd_awidth: integer:= 0; -- read fifo address width (same as fifo_wr_awidth but for read logic)
      use_toutSup: integer:= 0; -- >0 generate timeout Suppress signal whenever this slave is selected
      use_be: integer:= 1; -- use Byte Enable logic;  <=0 - Byte enable are always '1' (only full data transfres are used)
      user_full_transfer: integer:= 1; -- user side full data bus transfer, the parameter is used for  dwidth!= K*real_dwidth or real_dwidth!= K*dwidth
      USE_SeqAddr : integer:= 1 -- use seq_addr (=0 saves area but reduces the transfer speed for seq_addr)
      );
    port ( clk : in std_logic:= '0'; -- clock signal
      arst: in std_logic:= '0'; -- asynchronous reset (needed only for simulation)
      -- OPB signals
      OPB_ABus    : in std_logic_vector(0 to C_OPB_AWIDTH-1);
      OPB_BE      : in std_logic_vector(0 to C_OPB_DWIDTH/8-1);
      OPB_RNW     : in std_logic;
      OPB_select  : in std_logic;
      OPB_seqAddr : in std_logic;
      OPB_DBus    : in std_logic_vector(0 to C_OPB_DWIDTH-1);
      OPB_rst     : in std_logic:= '0';
      Sl_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
      Sl_errAck  : out std_logic;
      Sl_retry   : out std_logic;
      Sl_toutSup : out std_logic;
      Sl_xferAck : out std_logic;
      -- user signals
      sel: out std_logic; -- select - active when OPB_select and proper address
      ack: in std_logic; -- acknowlenge - slave is ready for data transfer
      retry: in std_logic:= '0'; -- generate retry signal
      rnw: out std_logic; -- read not write - same assert OPB_read not write
      adr: out std_logic_vector(0 to awidth-1); -- address bus
      drd: in std_logic_vector(0 to dwidth-1); -- data read by opb master (send by the uer device)
      dwr: out std_logic_vector(0 to dwidth-1); -- data written by the master (recieved by the user device)
      be: out std_logic_vector(0 to (dwidth+7)/8 -1); -- byte enable 
      seqAddr: out std_logic -- same as OPB_seqAddr
      );
  end component;
  
  -----------------------------------------------------------------------------------------------
  -- constants and signals declarations
  constant adr_space: integer:= conv_integer(c_highaddr - c_baseaddr + 1);
  constant awidth: integer:= log2(adr_space); -- address width when address spac is considered
  constant adr_lsb_size: integer:= log2( c_real_dwidth / C_FLASH_DWIDTH); -- LSBs of address which are generated locally due to multiply flash transfers for a single OPB transfer
  constant adr_ignore_size: integer:= log2(C_FLASH_DWIDTH/8); -- ignore LSBs of the address
  
  signal ABus: std_logic_vector(0 to awidth-1); -- flash module address bus
  signal BE: std_logic_vector(0 to c_real_dwidth/8 -1); -- Byte enable (excluding not used bits)
  signal arst: std_logic; -- asynchronous global set/ reset
  signal ack, sel, rnw, retry: std_logic; -- internal opb_slave signals
  signal drd, dwr: std_logic_vector(0 to c_real_dwidth-1); -- internal signals drd, dwr - form opb_slave, DBus- dwr for rd_before_wr=1
begin  
  sl: OPB_slave
  generic map (C_OPB_AWIDTH=> C_OPB_AWIDTH, C_OPB_DWIDTH=> C_OPB_DWIDTH,
    C_BASEADDR=> C_BASEADDR, C_HIGHADDR=> C_HIGHADDR, C_SPEEDRATE=> C_SPEEDRATE,
    DWIDTH=> C_REAL_DWIDTH, AWIDTH=> awidth, C_REAL_DWIDTH=> C_REAL_DWIDTH,
    fifo_wr_awidth=> 0, fifo_rd_awidth=> 0, use_toutSup=> 0,
    use_be=> 1, user_full_transfer=> 1, USE_SeqAddr=> 0)
  port map(clk=> OPB_clk, arst=> arst,
    -- OPB signals
    OPB_ABus=> OPB_ABus, OPB_BE=> OPB_BE, OPB_RNW=> OPB_RNW,
    OPB_select=> OPB_select, OPB_seqAddr=> OPB_seqAddr,
    OPB_DBus=> OPB_DBus, Sl_DBus=> Sl_DBus, Sl_errAck=> Sl_errAck,
    Sl_retry=> Sl_retry, Sl_toutSup=> Sl_toutSup, Sl_xferAck=> Sl_xferAck,
    -- user signals
    sel=> sel, ack=> ack, rnw=> rnw, adr=> ABus, retry=> retry,
    drd=> drd, dwr=> dwr, be=> be, seqAddr=> open);
  
    fl: flash2opb
    generic map(C_AWIDTH=> awidth - adr_ignore_size, C_DWIDTH=> c_real_dwidth, adr_lsb_size=> adr_lsb_size,
      C_FLASH_DWIDTH=> C_FLASH_DWIDTH, C_FLASH_AWIDTH=> C_FLASH_AWIDTH, C_WRITE_PULSE_NS=> C_WRITE_PULSE_NS,
      C_CLOCK_PERIOD_NS=> C_CLOCK_PERIOD_NS, C_ACCESS_TIME_NS=>C_ACCESS_TIME_NS)
    port map(arst=> arst, clk=> OPB_clk, 
      ABus=> ABus(0 to awidth-adr_ignore_size -1), DBus=> dwr, Sl_DBus=> drd,
  	  be=> be, rnw=> rnw, stb=> sel, ack=> ack, retry=> retry,
  	  -- external SRAM signals
      adr=> a, 
      data_i=> d_i, data_o=> d_o, data_t=> d_t,
      oen=> oen, wen=> wen, cen=> cen, busyn=> busyn);
  
  -- global set/reset logic - this component draws rst during chip configuration only (no external reset is used)
  global_reset: ROC
  port map (O=> arst);
  
  assert c_real_dwidth<=c_opb_dwidth
  report "opb_flash error. c_real_dwidth must be greater than c_opb_dwidth"
  severity error;
  assert c_real_dwidth>=c_flash_dwidth
  report "opb_flash error. c_real_dwidth must be greater or eq. to the c_flash_dwidth"
  severity error;
  assert adr_space*c_real_dwidth/c_opb_dwidth <= (2**c_flash_awidth) * c_flash_dwidth/8
  report "opb_flash error. The flash memory address space (HighAddress - BaseAddress) is too large in comparison to the actual flash size (flash_awidth, flash_dwidth)"
  severity error;
  
end arch;