-- Author:          Ernest Jamro
-- plik interface urzadzenia slave - kazde urzadzenie slave powinno korzystac z tego modulu
-- ta wersja urzadzenia jest poprawiona w stosunku do opb_slave
-- dlatego naley korzysta z tego moduu a nie ze starego moduo opb_slave

--//////////////////////////////////////////////////////////////////////
--//// Copyright (C) 2003 Authors                                   //// 
--////                                                              ////
--//// This source file may be used and distributed without         ////
--/// restriction provided that this copyright statement is not     ////
--//// removed from the file and that any derivative work contains  ////
--//// the original copyright notice and the associated disclaimer. ////
--////                                                              ////
--//// This source file is free software; you can redistribute it   ////
--//// and/or modify it under the terms of the GNU Lesser General   ////
--//// Public License as published by the Free Software Foundation; ////
--//// either version 2.1 of the License, or (at your option) any   ////
--//// later version.                                               ////
--////                                                              ////
--//// This source is distributed in the hope that it will be       ////
--//// useful, but WITHOUT ANY WARRANTY; without even the implied   ////
--//// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ////
--//// PURPOSE. See the GNU Lesser General Public License for more  ////
--//// details.                                                     ////
--//////////////////////////////////////////////////////////////////////


library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

entity OPB_slave1 is
  generic (
    C_OPB_AWIDTH : integer                   := 32;
    C_OPB_DWIDTH : integer                   := 32;
    C_BASEADDR   : std_logic_vector          := X"0000_0100";
    C_HIGHADDR   : std_logic_vector          := X"0000_01FF";
	  C_SPEEDRATE : integer := 0; -- the lower the speed the quicker (lower number of clock cycles - latency) 
	   -- and the lower occupied area however the lower clock frequency is required. 
     -- Therefore to increase the clock frequency this parameter should be increased
	  -- user interface data and address width
    USE_SeqAddr : integer:= 1; -- use seq_addr (=0 saves area but reduces the transfer speed for seq_addr)
	  DWIDTH : integer:= 8; -- real width of the data - the rest LSBs are always 0s
	  AWIDTH : integer:= 8 -- user address width
    );
  port ( clk : in std_logic; -- clock signal
    arst: in std_logic; -- asynchronous reset (needed only for simulation)
    -- OPB signals
    OPB_ABus    : in std_logic_vector(0 to C_OPB_AWIDTH-1);
    OPB_BE      : in std_logic_vector(0 to C_OPB_DWIDTH/8-1);
    OPB_RNW     : in std_logic;
    OPB_select  : in std_logic;
    OPB_seqAddr : in std_logic;
    OPB_DBus    : in std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sl_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sl_errAck  : out std_logic;
    Sl_retry   : out std_logic;
    Sl_toutSup : out std_logic;
    Sl_xferAck : out std_logic;
	  -- user signals
	  sel: out std_logic; -- select - active when OPB_select and proper address
	  ack: in std_logic; -- acknowlenge - slave is ready for data transfer
	  rnw: out std_logic; -- read not write - same assert OPB_read not write
	  adr: out std_logic_vector(0 to awidth-1); -- address bus
	  drd: in std_logic_vector(0 to dwidth-1); -- data read by opb master (send by the uer device)
	  dwr: out std_logic_vector(0 to dwidth-1); -- data written by the master (recieved by the user device)
	  be: out std_logic_vector(0 to dwidth/8 -1); -- byte enable 
	  seqAddr: out std_logic -- same as OPB_seqAddr
    );
end OPB_slave1;


architecture opb_slave_arch of OPB_slave1 is
  function log2(arg: integer) return integer is
    variable ret: integer:= -1;
  begin
   for i in 0 to 32 loop
	  if 2**i = arg then
		  ret:= i;
		  return i;
	  end if;
	end loop;
	assert false -- always assert when loop does not finish the function
	report "Address space is not a power of 2"
	severity failure;
	return -1;
end log2;
  constant adr_space: integer:= conv_integer(c_highaddr - c_baseaddr + 1);
  constant this_awidth: integer:= log2(adr_space); -- memory address width

  signal my_adr: std_logic; -- this device is selected (proper address)
  signal sel1, rnw1: std_logic; -- internal select signal
begin  
  
  adr_all: if awidth = c_opb_awidth generate -- the device is always selected
    my_adr<= '1';
    sel1<= OPB_select;
    rnw1<= OPB_RNW;
  end generate;
  
  an: if awidth < c_opb_awidth generate -- decode address space
    
    speed0: if c_speedrate=0 generate 
      my_adr<= '1' when OPB_ABus(0 to C_OPB_AWIDTH - awidth - 1) = C_BASEADDR(0 to C_OPB_AWIDTH - awidth - 1)
     else '0';
    
     sel1<= OPB_select and my_adr;
     rnw1<= OPB_rnw;
     be<= OPB_BE(0 to (dwidth-1)/8);
    end generate;

   speed1: if c_speedrate=1 generate
      process(arst, clk) begin
	      if arst='1' then my_adr<='0'; 
        elsif clk'event and clk='1' then
		      if OPB_ABus(0 to C_OPB_AWIDTH - awidth - 1) = C_BASEADDR(0 to C_OPB_AWIDTH - awidth - 1)
	          and (ack='0' or (OPB_seqAddr='1' and use_seqAddr>0)) then
		        my_adr<= '1';
		      else my_adr<='0';
	        end if;  
	      end if;
	    end process; 
    sel1<= OPB_select and my_adr;
    rnw1<= OPB_rnw;
    be<= OPB_BE(0 to (dwidth-1)/8);
  end generate; -- decode address space 
  
  speed2: if c_speedrate=2 generate
      process(arst, clk) begin
	      if arst='1' then sel1<= '0'; rnw1<= '0'; be<= (others=>'0');
	      elsif clk'event and clk='1' then
	        rnw1<= OPB_rnw;
          be<= OPB_BE(0 to (dwidth-1)/8);
          if OPB_ABus(0 to C_OPB_AWIDTH - awidth - 1) = C_BASEADDR(0 to C_OPB_AWIDTH - awidth - 1)
	          and (opb_select='1' and ack='0') then
		         sel1<= '1';
		      else 
             sel1<='0';
	        end if;  
	      end if;
	    end process; 
  end generate;  -- speed 2

end generate; -- decode address space 
  
  sel<= sel1;
  rnw<= rnw1;
  
  Sl_xferAck<= ack;
  Sl_errAck<= '0';
  Sl_retry<= '0';
  Sl_toutSup<= '0';

  gd: if c_opb_dwidth>dwidth generate -- different data width for OPB_bus and user bus
	 Sl_DBus(dwidth to c_opb_dwidth-1)<= (others=>'0');
  end generate;
  Sl_DBus(0 to dwidth-1)<= drd when sel1='1' and rnw1='1' else (others=>'0');
  dwr<= OPB_DBus(0 to dwidth-1);
  seqAddr<= OPB_seqAddr;
  adr<= OPB_ABus(c_opb_awidth-awidth to c_opb_awidth-1);

  
  assert this_awidth = awidth
   report "opb_slave1 module error. Different address width for awidth and address space"
   severity error;
   
  assert not (c_speedrate<0 or c_speedrate>=3)
    report "opb_slave1. The c_speedrate value must be in the range 0 to 2."
  	severity error;
	  
   
 end opb_slave_arch;