-- SRAM to OPB block
-- author: Erenst Jamro


library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;



entity sram2opb is
  generic (
  	C_AWIDTH: integer:= 18; -- number of address bits of the OPB bus connected to the SRAM= C_SRAM_AWIDTH - log2(C_DWIDTH/C_SRAM_DWIDTH)
    C_DWIDTH : integer:= 32; -- real width of the OPB data -bus  the rest MSBs are always 0s
	C_SRAM_DWIDTH: integer:= 16; -- external memory data width
	C_SRAM_AWIDTH: integer:= 19; -- external memory address width
	C_USE_SeqAddr: integer:= 0); -- use OPB_SeqAddr to speed up reading
  port (
  -- Global signals
    arst: in std_logic;
    clk : in std_logic;
    ABus    : in std_logic_vector(0 to c_awidth-1);
    DBus    : in std_logic_vector(0 to C_DWIDTH-1);
    Sl_DBus    : out std_logic_vector(0 to C_DWIDTH-1);
	-- internal signals diffrent from OPB
	be: in std_logic_vector(0 to c_dwidth/8 -1); -- byte enable
	we: in std_logic; -- write enable (when rd=1 and we='1' then read before wr command is started do not write when read is '1')
	rd: in std_logic; -- read command and device is selected
	xferAck: buffer std_logic; -- TRANSFER acknowledge
	ce_adr: in std_logic; -- address register clock enable
	seqAddr: in std_logic; -- increament address by 1 - valid only when C_USE_SeqAddr= 1
	-- external SRAM signals
    adr: out STD_LOGIC_VECTOR (0 to c_sram_awidth-1); -- address bus
    data_i: in STD_LOGIC_VECTOR (0 to c_sram_dwidth-1); -- data bus
	data_o: out STD_LOGIC_VECTOR (0 to c_sram_dwidth-1); -- data bus
	data_t: out STD_LOGIC; -- tri-state datap_o
    oen: out std_logic; -- output enable (active low)
    wen: out std_logic; -- write enable
    cen: out std_logic -- chip enable (always zero)	
    );
end entity sram2opb;

architecture sram2opb_arch of sram2opb is
component sram
    generic (adr_width: integer:= 19; -- width of the address bus
    	data_width: integer:= 16  -- width of the data bus
    	); 
    port ( -- user signals
	    clk: in STD_LOGIC; -- clock signal - every user signal is latched rising edge of clk
		arst: in std_logic; -- global asynchronous set reset 
        wr_rdn: in STD_LOGIC; -- write / not read
		adri: in STD_LOGIC_VECTOR (0 to adr_width-1); -- address from which data will be read or written
		datar: out STD_LOGIC_VECTOR (0 to data_width-1); -- data read form the sram
        dataw: in STD_LOGIC_VECTOR (0 to data_width-1); -- data to be written to the sram
		ce_adr: in std_logic; -- clock enable for address flip-flops (except the LSB which always always '1')
		ce_datar: in std_logic; -- clock enable for datar flip-flops
		ce_dataw: in std_logic_vector(0 to data_width/8 -1); -- clock enable for write flip-flops - seperate for each byte
        -- external SRAM signals
        adr: out STD_LOGIC_VECTOR (0 to adr_width-1); -- address bus 
		data_i: in STD_LOGIC_VECTOR (0 to data_width-1); -- data bus (input)
		data_o: out STD_LOGIC_VECTOR (0 to data_width-1); -- data bus (output)
		data_t: out STD_LOGIC; -- data bus (tri-state)
        oen: out std_logic; -- output enable (active low)
        wen: out std_logic; -- write enable
        cen: out std_logic -- chip enable (always zero)
    );
end component;
  constant k: integer:= c_dwidth/c_sram_dwidth; -- number of sram cycles for one OPB cycle
  signal wr_rdn: std_logic; -- write not read for sram
  signal adri, adr1: std_logic_vector(0 to c_sram_awidth-1); -- sram address (input flip-flop and output)
  --signal ce_adri: std_logic; -- clock enable for address 
  signal datar, dataw: std_logic_vector(0 to c_sram_dwidth-1); -- sram data read and write
  signal ce_datar: std_logic; -- clock enable for address and data read and write
  signal ce_dataw: std_logic_vector(0 to c_sram_dwidth/8-1); -- write filip flops clock enable
  signal datar_q, dataw_q: std_logic_vector(0 to c_sram_dwidth-1); -- flip flops needed to convert from sram to OPB data bus width
  signal be_q: std_logic_vector(0 to 1); -- byte enable flip-flops (only for the MSBs and rd_before_wr=1)
  signal valid_datar_q, valid_dataw_q: std_logic; -- valid data in the datar(w)_q registers
  signal valid_adr_q: std_logic; -- valid address in the adr_q flip flops (logic considered only for reading)
  signal valid_datar: std_logic; -- datar signal is valid 
  -- OPB signals
  
begin  
  mem: sram
     generic map (adr_width=> c_sram_awidth, data_width=> c_sram_dwidth)
     port map(clk=> clk, arst=> arst, 
        wr_rdn=> wr_rdn, adri=> adri, 
		datar=> datar, dataw=> dataw, ce_adr=> ce_adr, 
		ce_datar=> ce_datar, ce_dataw=> ce_dataw,
        -- external SRAM signals
        data_i=> data_i, data_o=> data_o,data_t=> data_t,
		adr=> adr, oen=> oen, wen=> wen, cen=> cen);
  
--------------------------------------------------------------------------
-- k= 2 - 2 SRAM transfers required for one OPB transfer
  -- for writes - first write the MSB and then the LSB, for reads - first read the MSB and then the LSB 
  -- consequently the LSB is transmitted very quickly in comparison to the LSB.
  k2: if k=2 generate
	-- dataw_q logic (at first write LSB and then MSB)
	d_wr_q: process(clk, arst) begin
	  if arst='1' then
		 dataw_q<= (others=>'0'); valid_dataw_q<= '0';
	  elsif clk'event and clk='1' then
		 dataw_q<= DBus(0 to c_sram_dwidth-1); -- MSBs of the OPB_DBus
		 if we='1' and (BE(0)='1' or BE(c_sram_dwidth/16)='1') and xferAck='1' then -- transfer data into the dataw_q buffer
			valid_dataw_q<= '1';
		 else -- transfer data out from the dataw_q buffer
		    valid_dataw_q<= '0';
		 end if;
	  end if;
	end process;
	-- dataw multiplexer
	dataw<= dataw_q when valid_dataw_q='1' else -- transfer MSB of the previous OPB transfer
	    DBus(c_sram_dwidth to c_dwidth-1); -- LSBs of the OPB

	-- datar_q logic (first read MSB and then LSB)
	d_rd_q: process(clk, arst) begin
	  if arst='1' then
		 datar_q<= (others=>'0');
	  elsif clk'event and clk='1' then
		datar_q<= datar; -- copy the MSBs to the read buffer
      end if;
	end process;
	
	Sl_DBus<= datar_q & datar;
	ce_dataw<= (others=>'1'); -- always '1'
	ce_datar<= '1';
  wr_rdn<= valid_dataw_q or (we and (BE(c_dwidth/8-1) or BE(c_sram_dwidth/16+1))); -- MSBs (valid_dataw_q) or  LSBs write (rst)
	xferAck<= ( not BE(0) and not BE(c_dwidth/16-1) and  valid_datar ) -- only a 16-bits LSBs read
         or valid_datar_q -- full (32-bit read) (for 16-bit MSB read also 32 -bit transfer is used - this allows for one mux saving)
        when rd='1'
        else not valid_dataw_q when we='1' -- for writes
	    else '0';
	
  -- do not use seqAddr		 
 seqAddr0: if c_use_seqAddr<= 0 generate
	  process(arst, clk) begin
	    if arst='1' then valid_adr_q<= '0'; valid_datar<= '0';
	    elsif clk'event and clk='1' then
		  if rd='1' and xferAck='0' and valid_dataw_q='0' then
		    valid_adr_q<= '1';
		  else 
		    valid_adr_q<='0';
		  end if;
		  valid_datar<= valid_adr_q and rd and not xferAck;
		  valid_datar_q<= valid_datar and rd and not xferAck;
	    end if;
	  end process;

	  adri(c_sram_awidth-1)<=  -- the LSB of the address
	    not valid_dataw_q when we='1' or valid_dataw_q='1' else -- first write the LSB (adr(LSB)=1)
		-- read logic (below)
		valid_adr_q -- when the first read ( valid_adr_q= 0 when first read = '1' when the 2-nd read)
		or not  (BE(0) or BE(c_sram_dwidth/16)); -- first read MSB (0) then the LSB (1), if MSB BE='1' then read only LSB
	  -- adr(LSB)<= '1' for the LSB, <= '0' for the MSB
	  adri(0 to c_sram_awidth-2)<= ABus;
  end generate; -- useSeqAddr= 0
  

    -- use seqAddr		 
  seqAddr1: if c_use_seqAddr> 0 generate
	 process(arst, clk) begin
	   if arst='1' then valid_adr_q<= '0'; valid_datar<= '0';
	   elsif clk'event and clk='1' then
		 if rd='1' and valid_dataw_q='0' and (xferAck='0' or seqAddr='1') then
		   valid_adr_q<= '1';
		 else 
		   valid_adr_q<='0';
		 end if;
		 valid_datar<= valid_adr_q and rd and (not xferAck or SeqAddr);
		 valid_datar_q<= valid_datar and rd and not xferAck;
	   end if;
	 end process;

	
    adri(c_sram_awidth-1)<=  -- the LSB of the address
	    not valid_dataw_q when we='1' or valid_dataw_q='1' else -- first write the LSB (adr(LSB)=1)
		-- read logic (below)
		valid_adr_q -- when the first read ( valid_adr_q= 0 when first read (MSB), = '1' when the 2-nd read (LSB))
		or not  (BE(0) or BE(c_sram_dwidth/16)) when seqAddr='0' or valid_adr_q='0' -- if MSB BE='1' then read only the LSBs (start from 1)
        else not adr1(c_sram_awidth-1); -- increase address by 1 
	  -- adr(LSB)<= '1' for the LSB, <= '0' for the MSB
    
        -- internal address multiplexer
    adri(0 to c_sram_awidth-2)<= adr1(0 to c_sram_awidth-2) when valid_dataw_q='1' else -- keep old value when write buffer is valid
       Abus when valid_adr_q='0' or SeqAddr='0' or rd='0' -- initialise address
       else adr1(0 to c_sram_awidth-2) when adr1(c_sram_awidth-1)= '0' -- increase only the LSB
       else adr1(0 to c_sram_awidth-2) + 1; -- increase the whole address

    -- adr1 flip flops (same as adr but internal)
    process(arst, clk) begin
      if arst='1' then adr1<= (others=>'0');
      elsif clk'event and clk='1' then
        adr1<= adri;
      end if;
    end process;

    
  end generate; -- useSeqAddr= 1

  
  
 end generate; -- k=2	
 
--------------------------------------------------------------------------------------------------------- 
 --------------------------------------------------------------------------------------------------------
 
 k1: if k=1 generate -- the same sram and OPB data widths
	dataw<= DBus;
	Sl_DBus<= datar;
	wr_rdn<= we; 
	ce_dataw<= BE;
	ce_datar<= '1';
    xferAck<= valid_datar when rd='1' -- for reads
       else we;  -- for writes always ready
	
  -- do not use seqAddr		 
  seqAddr0: if c_use_seqAddr<= 0 generate
    adri<= ABus;
      
    process(arst, clk) begin
	  if arst='1' then valid_adr_q<= '0'; valid_datar<= '0';
	  elsif clk'event and clk='1' then
		if rd='1' and xferAck='0' then
		  valid_adr_q<= '1';
		else 
		  valid_adr_q<='0';
		end if;
		valid_datar<= valid_adr_q and rd and not xferAck;
	  end if;
	end process;
	
  end generate; -- useSeqAddr= 0
    
  -- use seqAddr		 
  seqAddr1: if c_use_seqAddr> 0 generate
	process(arst, clk) begin
	  if arst='1' then valid_adr_q<= '0'; valid_datar<= '0';
	  elsif clk'event and clk='1' then
		if rd='1' and (xferAck='0' or SeqAddr='1') then
		  valid_adr_q<= '1';
		else 
		  valid_adr_q<='0';
		end if;
		valid_datar<= valid_adr_q and rd and (not xferAck or SeqAddr);
	  end if;
	end process;
    -- internal address multiplexer
    adri<= ABus when valid_adr_q= '0' or seqAddr='0' or rd='0'
       else  adr1 + 1; -- increase address by 1
    -- adr1 flip flops (same as adr but internal)
    process(arst, clk) begin
      if arst='1' then adr1<= (others=>'0');
      elsif clk'event and clk='1' then
        adr1<= adri;
      end if;
    end process;
              
	process(arst, clk) begin
	  if arst='1' then valid_adr_q<= '0'; valid_datar<= '0';
	  elsif clk'event and clk='1' then
		if rd='1' and (xferAck='0' or SeqAddr='1') then
		  valid_adr_q<= '1';
		else 
		  valid_adr_q<='0';
		end if;
		valid_datar<= valid_adr_q and rd and (not xferAck or SeqAddr);
	  end if;
	end process;    
		
  end generate; -- useSeqAddr= 1  
  

end generate; -- k=1	 

end sram2opb_arch;