-- a model of SRAM memory for simulation purposes only
-- this model reads a binary file for memory initialisation
-- after a difined time (1ms) RAM contents will be writen to a file 
-- for file operation data width must be 16-bits 

library IEEE;	 
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;

entity sram_model is
	generic (data_width: integer:= 16; -- width of data bus
	    adr_width: integer:= 4); -- width of the address bus
	port(oen : in std_logic; -- output enable (active low)
		cen: in std_logic; -- chip enable
		wen: in std_logic; -- write enable
	  	adr: in std_logic_vector(adr_width-1 downto 0); -- address bus
   		data: inout std_logic_vector(0 to data_width-1)); -- bidirectional data bus
end sram_model;

architecture sram_model_arch of sram_model is  

type BIT_VECTOR_FILE is file of integer;

constant Data_File_In : string := "sramin.bin";  	-- nazwa pliku z danymi inicjujacymi pamiec	
constant Data_File_Out : string := "sramout.bin";  -- nazwa pliku do zapisu zawartoci pamieci
constant mem_write_time: time:= 1ms; -- czas po ktrym zawarto pamici sram zostanie zapisana do pliku Data_File_Out
type mem_data_type is array (0 TO 2**adr_width-1) of std_logic_vector(0 to data_width-1); 
 
 -- this function converts std_logic_vector to integer
function conv_int(arg: std_logic_vector) return integer is --unsigned input data
    variable ret: integer;
  begin
	ret:= 0;
  	for i in arg'high downto arg'low loop
		ret:= ret*2;
		if arg(i)= '1' then ret:= ret+1; end if;
	end loop;
    return ret;
  end;

 
procedure RAM_Init(signal fRAM : out mem_data_type ) is	 
  constant no_reads: integer:= 32/data_width; -- liczba lokacji adresowych na jedno czytanie z pliku (32 bity)
  Variable Data_In : integer;
  Variable din_unsigned: unsigned(0 to 31);
  Variable Data_Tmp : std_logic_vector(0 to data_width-1);
  variable Adres, I, J, K, L : NATURAL := 0;
  file Data_In_File : BIT_VECTOR_FILE;   

begin
   	FILE_OPEN ( Data_In_File, Data_File_In, READ_MODE); 
		-- jednorazowo czytane sa 4 bajty (integer)
		for  Adres in 0 to ((2 ** adr_width)/no_reads) - 1 loop
			READ ( Data_In_File, Data_In );
			din_unsigned:= conv_unsigned(Data_In, 32);
			for J in 0 to no_reads -1 loop -- dla kazdej lokacji adresowej
				I:= 0; 
				for k in data_width/8 -1 downto 0 loop	 -- dla kazdego bajtu
				   L:= (no_reads-1-j)*data_width+8*K;
				   data_tmp(8*I to 8*I +7):= 
				     std_logic_vector( din_unsigned(L to L+7));
				   I:= I+1;
				end loop;   
				fRAM(Adres*no_reads+J)<= data_tmp;
			end loop;
			
		end loop;
		FILE_CLOSE ( Data_In_File );
end;

-- procedura zapisujaca zawartosc pamieci do pliku
procedure RAM_Save(signal fRAM : in mem_data_type) is	 

	variable Data_out : integer; 						
	variable Adres, I, J : NATURAL := 0;
	file Data_Out_File : BIT_VECTOR_FILE;   

begin
   	FILE_OPEN ( Data_Out_File, Data_File_Out, WRITE_MODE); 
		-- jednorazowo zapisywane sa 4 bajty (integer)
		for  Adres in 0 to ((2 ** adr_width)/2) - 1 loop
			Data_out := 0;
			for J in 1 downto 0 loop
				for I in data_width-1 downto 0 loop 
					if fRAM(Adres * 2 + J)(I) = '0' then
						Data_out := Data_out + Data_out;
					elsif fRAM(Adres * 2 + J)(I) = '1' then	
						Data_out := Data_out + Data_out + 1;
					end if;
				end loop;
			end loop;
			WRITE ( Data_Out_File, Data_out );
		end loop;
		FILE_CLOSE ( Data_Out_File );
end;

 signal mem_data: mem_data_type;

begin
 
 -- ladowanie danych do pamieci
process begin
   wait for mem_write_time; -- czas po ktorym dane beda zapisywane
   RAM_SAVE(mem_data);
   wait for 1000ms;
 end process; 
 
 -- controls data HiZ
process(oen, cen, wen, adr) begin
 if oen='0' and cen='0' and wen='1' then
	data<= mem_data(conv_int(adr));
 else
	data<= (others=>'Z');
 end if;
end process;

 -- writing data to memory
process(wen, adr, data)
variable Initialize : integer := 0;			    -- flaga inicjalizacji
begin
	if Initialize=0 then						-- musimy zainicjowac pamiec 
	    RAM_Init( mem_data );						-- odczytujemy z pliku
		Initialize := 1;
   	elsif wen='0' then
	mem_data(conv_int(adr))<= data;
  end if;
end process;
--


end sram_model_arch;